import pandas as pd

from uk_postcodes_parsing.fix import fix
from uk_postcodes_parsing import ukpostcode
from uk_postcodes_parsing.ukpostcode import (
    parse_from_corpus,
    Postcode,
)


def test_fix():
    # Trims postcode
    assert fix(" SW1A 2AA ") == "SW1A 2AA"
    # Upper case
    assert fix(" Sw1A 2aa ") == "SW1A 2AA"
    # Fixes spacing
    assert fix(" Sw1A2aa ") == "SW1A 2AA"
    assert fix(" Sw1A    2aa ") == "SW1A 2AA"
    # Return original if not fixble
    assert fix(" 1A2aa ") == " 1A2aa "

    ## Outward code
    # Fixes LN format
    assert fix("01 OAA") == "O1 0AA"
    assert fix("SO OAA") == "S0 0AA"

    # Fixes L?? format
    assert fix("0W1 OAA") == "OW1 0AA"
    # ambiguous
    assert fix("S01 OAA") == "S01 0AA"
    assert fix("SO1 OAA") == "SO1 0AA"
    assert fix("SWO OAA") == "SWO 0AA"
    assert fix("SW0 OAA") == "SW0 0AA"

    # Fixes LNN?
    assert fix("0W1A OAA") == "OW1A 0AA"
    assert fix("S01A OAA") == "SO1A 0AA"
    assert fix("SWOA OAA") == "SW0A 0AA"
    # ambiguous
    assert fix("SW10 OAA") == "SW10 0AA"
    assert fix("SW1O OAA") == "SW1O 0AA"

    ## Inward code
    # first character
    assert fix(" SW1A OAA") == "SW1A 0AA"
    # second character
    assert fix("SW1A 20A") == "SW1A 2OA"
    # third character
    assert fix("SW1A 2A0") == "SW1A 2AO"
    # 1 <=> I
    assert fix("SWIA 2AA") == "SW1A 2AA"
    assert fix("1W1A 2AA") == "IW1A 2AA"


def test_parsing():
    corpus = "this is a check to see if we can get post codes liek thia ec1r   1ub , and that e3 4ss. But also eh16 50y and ei412"
    lst = parse_from_corpus(corpus)
    assert lst == [
        Postcode(
            original="ec1r   1ub",
            postcode="EC1R 1UB",
            incode="1UB",
            outcode="EC1R",
            area="EC",
            district="EC1",
            sub_district="EC1R",
            sector="EC1R 1",
            unit="UB",
        ),
        Postcode(
            original="e3 4ss",
            postcode="E3 4SS",
            incode="4SS",
            outcode="E3",
            area="E",
            district="E3",
            sub_district=None,
            sector="E3 4",
            unit="SS",
        ),
    ]

    corpus = "this is a check to see if we can get post codes liek thia ec1r 1ub , and that e34ss. But also eh16 50y and ei412. followed by ehi6 50y"
    lst = parse_from_corpus(corpus, attempt_fix=True)
    assert lst == [
        Postcode(
            original="ec1r 1ub",
            postcode="EC1R 1UB",
            incode="1UB",
            outcode="EC1R",
            area="EC",
            district="EC1",
            sub_district="EC1R",
            sector="EC1R 1",
            unit="UB",
        ),
        Postcode(
            original="e34ss",
            postcode="E3 4SS",
            incode="4SS",
            outcode="E3",
            area="E",
            district="E3",
            sub_district=None,
            sector="E3 4",
            unit="SS",
        ),
        Postcode(
            original="eh16 50y",
            postcode="EH16 5OY",
            incode="5OY",
            outcode="EH16",
            area="EH",
            district="EH16",
            sub_district=None,
            sector="EH16 5",
            unit="OY",
        ),
        Postcode(
            original="ehi6 50y",
            postcode="EH16 5OY",
            incode="5OY",
            outcode="EH16",
            area="EH",
            district="EH16",
            sub_district=None,
            sector="EH16 5",
            unit="OY",
        ),
    ]


def test_parsing_detailed():
    df = pd.read_csv("tests/data/postcode_parse_test.csv")
    df = df.where(pd.notnull(df), None)

    parsed_postcodes = df["Postcode"].apply(ukpostcode.parse)

    for i, parsed_postcode in enumerate(parsed_postcodes):
        assert parsed_postcode.outcode == df.iloc[i][".outcode"]
        assert parsed_postcode.incode == df.iloc[i][".incode"]
        assert parsed_postcode.area == df.iloc[i][".area"]
        assert parsed_postcode.district == df.iloc[i][".district"]
        assert parsed_postcode.sub_district == df.iloc[i][".subDistrict"]
        assert parsed_postcode.sector == df.iloc[i][".sector"]
        assert parsed_postcode.unit == df.iloc[i][".unit"]


def test_sort_by_fix_distance():
    corpus = "this EC1r 1ub followed by to one, ecir iub e0 i00"
    lst = parse_from_corpus(corpus, attempt_fix=True)
    assert sorted(lst, reverse=True) == [
        Postcode(
            original="ecir iub",
            postcode="EC1R 1UB",
            incode="1UB",
            outcode="EC1R",
            area="EC",
            district="EC1",
            sub_district="EC1R",
            sector="EC1R 1",
            unit="UB",
        ),
        Postcode(
            original="ec1r 1ub",
            postcode="EC1R 1UB",
            incode="1UB",
            outcode="EC1R",
            area="EC",
            district="EC1",
            sub_district="EC1R",
            sector="EC1R 1",
            unit="UB",
        ),
        Postcode(
            original="e0 i00",
            postcode="E0 1OO",
            incode="1OO",
            outcode="E0",
            area="E",
            district="E0",
            sub_district=None,
            sector="E0 1",
            unit="OO",
        ),
        Postcode(
            original="to one",
            postcode="T0 0NE",
            incode="0NE",
            outcode="T0",
            area="T",
            district="T0",
            sub_district=None,
            sector="T0 0",
            unit="NE",
        ),
    ]



"""
ukpostcode.py: main module for parsing UK postcodes from text.
"""
import re
import logging
from dataclasses import dataclass, field
from typing import Union, List, Optional

import requests

from uk_postcodes_parsing.postcode_utils import (
    is_valid,
    check_with_postcode_io,
    to_normalised,
    to_outcode,
    to_incode,
    to_area,
    to_sector,
    to_unit,
    to_district,
    to_sub_district,
)
from uk_postcodes_parsing.fix import fix

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("uk-postcodes-parsing")

# Test for a valid postcode embedded in text
POSTCODE_CORPUS_REGEX = re.compile(r"[a-z]{1,2}\d[a-z\d]?\s*\d[a-z]{2}", re.I)
FIXABLE_POSTCODE_CORPUS_REGEX = re.compile(
    r"[a-z01]{1,2}[0-9oi][a-z\d]?\s*[0-9oi][a-z01]{2}"
)


@dataclass(order=True)
class Postcode:
    """Class to hold the parsed postcode.
    Constructor arguments:
        original (str): The raw (original) string of the postcode.
        postcode (str): The postcode as a string.
        incode (str): The inward code (the first 3 characters) of the postcode.
        outcode (str): The outward code (the last 4 characters) of the postcode.
        area (str): The area of the postcode.
        district (str): The district of the postcode.
        sub_district (str): The sub-district of the postcode.
        sector (str): The sector of the postcode.
        unit (str): The unit of the postcode.

    Additional attributes:
        verified_against_postcode_io (bool): Whether the postcode was successfully verified against
            postcode [api.postcodes.io/postcodes/](https://postcodes.io/) which uses the
            [ONS Postcode Directory](https://geoportal.statistics.gov.uk/datasets/489c152010a3425f80a71dc3663f73e1/about).
        fix_distance (int): The number of characters that the postcode string was corrected by the `fix` function during parsing.
    """

    # Calculate post initialization
    verified_against_postcode_io: bool = field(init=False)
    fix_distance: int = field(init=False)
    # raw text
    original: str
    # The rest of the fields are parsed from the postcode using regex
    postcode: str
    incode: str
    outcode: str
    area: str
    district: str
    sub_district: Union[str, None]
    sector: str
    unit: str

    def __post_init__(self):
        """calculate class attributes after initialization.
        - `verified_against_postcode_io` (bool)
        - `fix_distance` (int): the "edit distance" between the raw postcode and final postcode
        """
        # Convert raw (original) string in postcode format
        original = self.original.upper().strip().replace(r"\s+", "")
        inward = original[-3:].strip()
        outward = original[:-3].strip()
        formatted = f"{outward} {inward}"

        fix_distance = sum(c1 != c2 for c1, c2 in zip(formatted, self.postcode))
        self.fix_distance = fix_distance

        try:
            self.verified_against_postcode_io = check_with_postcode_io(self.postcode)
        except requests.exceptions.RequestException:
            logger.error("Failed to check postcode against postcode.io")

    def __eq__(self, other):
        """Ignore verified_against_postcode_io and fix_distance."""
        return (
            self.original == other.original
            and self.postcode == other.postcode
            and self.incode == other.incode
            and self.outcode == other.outcode
            and self.area == other.area
            and self.district == other.district
            and self.sub_district == other.sub_district
            and self.sector == other.sector
            and self.unit == other.unit
        )


def _parse(postcode: str) -> dict:
    """Internal function to parse postcode string using `uk_postcodes_parsing.postcode_utils`

    Args:
        postcode (str): Postcode to parse. E.g. "EC1R 1UB".
    Returns:
        dict: Parsed postcode.
    """
    if postcode.upper().startswith("NPT"):  # Edge case logging
        logger.warning("Found 'NPT' Newport postcode discontinued in 1984.")
    if not is_valid(postcode):
        return None
    return {
        "postcode": to_normalised(postcode),
        "incode": to_incode(postcode),
        "outcode": to_outcode(postcode),
        "area": to_area(postcode),
        "district": to_district(postcode),
        "sub_district": to_sub_district(postcode),
        "sector": to_sector(postcode),
        "unit": to_unit(postcode),
    }


def parse(postcode, attempt_fix=True) -> Optional[Postcode]:
    """Parse a postcode

    Args:
        postcode (str): Postcode to parse. E.g. "EC1R 1UB".
        attempt_fix (bool): Attempt to fix postcodes. Defaults to True.
    Returns:
        Postcode: Parsed postcode.
    """
    if is_valid(postcode):
        return Postcode(**_parse(postcode), original=postcode)
    if attempt_fix:
        fixed = fix(postcode)
        if is_valid(fixed):
            logger.info("Postcode Fixed: '%s' => '%s'", postcode, fixed)
            return Postcode(**_parse(fixed), original=postcode)
        logger.error("Unable to fix postcode")
    logger.error("Failed to parse postcode")
    return None


def parse_from_corpus(text: str, attempt_fix=False) -> List[str]:
    """Parse postcodes from a text corpus

    Args:
        text (str): Text corpus. E.g. "The postcode could be EC1R 1UB or EC1R IUB"
        attempt_fix (bool): Attempt to fix postcodes. Defaults to False.
    Returns:
        List[Postcode]: List of parsed postcodes.

    """
    if attempt_fix:
        postcodes = re.findall(FIXABLE_POSTCODE_CORPUS_REGEX, text.lower())
    else:
        postcodes = re.findall(POSTCODE_CORPUS_REGEX, text.lower())
    logger.info("Found %d postcodes in corpus", len(postcodes))
    return list(map(parse, postcodes))



import re

FIXABLE_REGEX = re.compile(
    r"^\s*[a-z01]{1,2}[0-9oi][a-z\d]?\s*[0-9oi][a-z01]{2}\s*$", re.I
)


def fix(s: str) -> str:
    """Attempts to fix a given postcode

    Args:
        s (str): The postcode to fix
    Returns:
        str: The fixed postcode
    """
    if not FIXABLE_REGEX.match(s):
        return s
    s = s.upper().strip().replace(r"\s+", "")
    inward = s[-3:].strip()
    outward = s[:-3].strip()
    return f"{coerce_outcode(outward)} {coerce_incode(inward)}"


def to_letter(char: str) -> str:
    """
    Convert a number to a letter if possible. For example,
        "0" => "O"
        "1" => "I"

    Args:
        c (str): The number to convert
    Returns:
        str: The letter
    """
    return {
        "0": "O",
        "1": "I",
    }.get(char, char)


def to_number(char: str) -> str:
    """
    Convert a letter to a number if possible. For example,
        "O" => "0"
        "I" => "1"

    Args:
        c (str): The letter to convert
    Returns:
        str: The number
    """
    return {
        "O": "0",
        "I": "1",
    }.get(char, char)


def coerce(pattern: str, string: str) -> str:
    """
    Coerce a given string into a pattern. For e.g.
        pattern: "LN" => signifies a letter followed by a number
        input: "01" => changes to "O1". Where the first 0 should be a letter so it
            is coverted from 0 to O. The Second 1 is already correct so no change.

    Args:
        pattern (str): The pattern to coerce the string into
        string (str): The string to coerce

    Returns:
        str: The coerced string (with the pattern applied to it)
    """
    return "".join(
        to_number(c) if target == "N" else to_letter(c) if target == "L" else c
        for c, target in zip(string, pattern)
    )


def coerce_outcode(i: str) -> str:
    """Coerce outcode"""
    if len(i) == 2:
        return coerce("LN", i)
    elif len(i) == 3:
        return coerce("L??", i)
    elif len(i) == 4:
        return coerce("LLN?", i)
    else:
        return i


def coerce_incode(i: str) -> str:
    """Coerce incode"""
    return coerce("NLL", i)



"""
postcode_utils.py: Utilities for working with postcodes.
"""
import re
from typing import Union

import requests

# Tests for district
DISTRICT_SPLIT_REGEX = re.compile(r"^([a-z]{1,2}\d)([a-z])$", re.I)

# Tests for the unit section of a postcode
UNIT_REGEX = re.compile(r"[a-z]{2}$", re.I)

# Tests for the inward code section of a postcode
INCODE_REGEX = re.compile(r"\d[a-z]{2}$", re.I)

# Tests for the outward code section of a postcode
OUTCODE_REGEX = re.compile(r"^[a-z]{1,2}\d[a-z\d]?$", re.I)

# Tests for a valid postcode
POSTCODE_REGEX = re.compile(r"^[a-z]{1,2}\d[a-z\d]?\s*\d[a-z]{2}$", re.I)

# Tests for the area section of a postcode
AREA_REGEX = re.compile(r"^[a-z]{1,2}", re.I)


def sanitize(string: str) -> str:
    """Sanitizes a string by removing whitespace and converting to uppercase.
    Args:
        s (str): The string to sanitize
    Returns:
        str: The sanitized string
    """
    return string.replace(" ", "").upper()


def is_valid(postcode: str) -> bool:
    """
    Checks if a postcode is valid using the `POSTCODE_REGEX`.
    Args:
        postcode (str): The postcode to check

    Returns:
        bool: True if the postcode is valid, False otherwise
    """
    return re.match(POSTCODE_REGEX, postcode) is not None


def is_valid_outcode(outcode: str) -> bool:
    """
    Checks if a string representing an outcode is valid using the `OUTCODE_REGEX`.
    Args:
        outcode (str): The postcode to check
    Returns:
        bool: True if the postcode is valid, False otherwise
    """
    return re.match(OUTCODE_REGEX, outcode) is not None


def check_with_postcode_io(postcode: str) -> bool:
    """Check if the postcode is valid with postcodes.io

    Args:
        postcode (str): The postcode to check

    Returns:
        bool: True if the postcode is valid, False otherwise
    """
    response = requests.get(f"https://api.postcodes.io/postcodes/{postcode}").json()
    return response["status"] == 200


def to_normalised(postcode: str) -> Union[str, None]:
    """
    Normalises a postcode by removing whitespace, converting to uppercase, and formatting.
    Args:
        postcode (str): The postcode to normalise
    Returns:
        str: The normalised postcode
    """
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    incode = to_incode(postcode)
    return None if incode is None else f"{outcode} {incode}"


def to_outcode(postcode: str) -> Union[str, None]:
    """Extract the outcode from a postcode string.
    Args:
        postcode (str): The postcode to extract the outcode from
    Returns:
        str: The outcode
    """
    if not is_valid(postcode):
        return None
    return re.sub(INCODE_REGEX, "", sanitize(postcode))


def to_incode(postcode: str) -> Union[str, None]:
    """Extract the incode from a postcode string.
    Args:
        postcode (str): The postcode to extract the incode from
    Returns:
        str: The incode
    """
    if not is_valid(postcode):
        return None
    incode = re.findall(INCODE_REGEX, sanitize(postcode))
    return incode[0] if incode else None


def to_area(postcode: str) -> Union[str, None]:
    """Extract the area from a postcode string.
    Args:
        postcode (str): The postcode to extract the area from
    Returns:
        str: The area
    """
    if not is_valid(postcode):
        return None
    area = re.findall(AREA_REGEX, sanitize(postcode))
    return area[0] if area else None


def to_sector(postcode: str) -> Union[str, None]:
    """Extract the sector from a postcode string.
    Args:
        postcode (str): The postcode to extract the sector from
    Returns:
        str: The sector
    """
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    incode = to_incode(postcode)
    return None if incode is None else f"{outcode} {incode[0]}"


def to_unit(postcode: str) -> Union[str, None]:
    """Extract the unit from a postcode string.
    Args:
        postcode (str): The postcode to extract the unit from
    Returns:
        str: The unit
    """
    if not is_valid(postcode):
        return None
    unit = re.findall(UNIT_REGEX, sanitize(postcode))
    return unit[0] if unit else None


def to_district(postcode: str) -> Union[str, None]:
    """Extract the district from a postcode string.
    Args:
        postcode (str): The postcode to extract the district from
    Returns:
        str: The district
    """
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    district = re.match(DISTRICT_SPLIT_REGEX, outcode)
    return district[1] if district else outcode


def to_sub_district(postcode: str) -> Union[str, None]:
    """Extract the sub-district from a postcode string.
    Args:
        postcode (str): The postcode to extract the sub-district from
    Returns:
        str: The sub-district
    """
    outcode = to_outcode(postcode)
    if outcode is None:
        return None
    split = re.match(DISTRICT_SPLIT_REGEX, outcode)
    return None if split is None else outcode
